# Get current project version from Git tags or changelog.

import re
from contextlib import suppress
from pathlib import Path

from pdm.backend.hooks.version import SCMVersion, Version, default_version_formatter, get_version_from_scm

_root = Path(__file__).parent.parent
_changelog = _root / "CHANGELOG.md"
_changelog_version_re = re.compile(r"^## \[(\d+\.\d+\.\d+)\].*$")
_default_scm_version = SCMVersion(Version("0.0.0"), None, False, None, None)  # noqa: FBT003


def get_version() -> str:
    scm_version = get_version_from_scm(_root) or _default_scm_version
    if scm_version.version <= Version("0.1"):  # Missing Git tags?
        with suppress(OSError, StopIteration):  # noqa: SIM117
            with _changelog.open("r", encoding="utf8") as file:
                match = next(filter(None, map(_changelog_version_re.match, file)))
                scm_version = scm_version._replace(version=Version(match.group(1)))
    return default_version_formatter(scm_version)


if __name__ == "__main__":
    print(get_version())



from __future__ import annotations

import logging
import sys
from dataclasses import dataclass, fields
from dataclasses import field as dataclass_field
from pathlib import Path
from typing import TYPE_CHECKING, Any
from typing import Annotated as An

from typing_extensions import Doc

# DUE: EOL 3.10: Replace block with line 2.
if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping


_logger = logging.getLogger(__name__)


class Unset:
    """A sentinel value for unset configuration options."""

    def __init__(
        self,
        key: An[str, Doc("TOML key.")],
        transform: An[str | None, Doc("Name of the method to call to transform the config value.")] = None,
    ) -> None:
        self.key: An[str, Doc("TOML key.")] = key
        self.name: An[str, Doc("Transformed key name.")] = key.replace("-", "_").replace(".", "_")
        self.transform: An[str | None, Doc("Name of the method to call to transform the config value.")] = transform

    def __bool__(self) -> bool:
        """An unset value always evaluates to False."""
        return False

    def __repr__(self) -> str:
        return f"<Unset({self.name!r})>"

    def __str__(self) -> str:
        # The string representation is used in the CLI, to show the default values.
        return f"`{self.key}` config-value"


def config_field(
    key: An[str, Doc("Key within the config file.")],
    transform: An[str | None, Doc("Name of transformation method to apply.")] = None,
) -> An[Unset, Doc("Configuration field.")]:
    """Create a dataclass field with a TOML key."""
    return dataclass_field(default=Unset(key, transform=transform))


# DUE: EOL 3.9: Remove block.
_dataclass_opts: dict[str, bool] = {}
if sys.version_info >= (3, 10):
    _dataclass_opts["kw_only"] = True


# DUE: EOL 3.9: Replace `**_dataclass_opts` with `kw_only=True` within line.
@dataclass(**_dataclass_opts)
class Config:
    """Configuration for the insiders project."""

    prefix: An[list[str] | Unset, Doc("The prefix for Yore comments.")] = config_field("prefix")  # noqa: RUF009
    diff_highlight: An[str | Unset, Doc("The command to highlight diffs.")] = config_field("diff.highlight")  # noqa: RUF009

    @classmethod
    def _get(
        cls,
        data: An[Mapping[str, Any], Doc("Data to get value from.")],
        *keys: An[str, Doc("Keys to access nested dictionary.")],
        default: An[Unset, Doc("Default value if key is not found.")],
        transform: An[Callable[[Any], Any] | None, Doc("Transformation function to apply to the value.")] = None,
    ) -> An[Any, Doc("Value from the nested dictionary.")]:
        """Get a value from a nested dictionary."""
        for key in keys:
            if key not in data:
                return default
            data = data[key]
        if transform:
            return transform(data)
        return data

    @classmethod
    def from_data(
        cls,
        data: An[Mapping[str, Any], Doc("Data to load configuration from.")],
    ) -> An[Config, Doc("Loaded configuration.")]:
        """Load configuration from data."""
        # Check for unknown configuration keys.
        field_keys = [field.default.key for field in fields(cls)]  # type: ignore[union-attr]
        unknown_keys = []
        for top_level_key, top_level_value in data.items():
            if isinstance(top_level_value, dict):
                for key in top_level_value.keys():  # noqa: SIM118
                    final_key = f"{top_level_key}.{key}"
                    if final_key not in field_keys:
                        unknown_keys.append(final_key)
            elif top_level_key not in field_keys:
                unknown_keys.append(top_level_key)
        if unknown_keys:
            _logger.warning(f"Unknown configuration keys: {', '.join(unknown_keys)}")

        # Create a configuration instance.
        return cls(
            **{
                field.name: cls._get(
                    data,
                    *field.default.key.split("."),  # type: ignore[union-attr]
                    default=field.default,  # type: ignore[arg-type]
                    transform=getattr(cls, field.default.transform or "", None),  # type: ignore[union-attr]
                )
                for field in fields(cls)
            },
        )

    @classmethod
    def from_file(
        cls,
        path: An[str | Path, Doc("Path to the configuration file.")],
    ) -> An[Config, Doc("Loaded configuration.")]:
        """Load configuration from a file."""
        with open(path, "rb") as file:
            return cls.from_data(tomllib.load(file))

    @classmethod
    def from_pyproject(
        cls,
        path: An[str | Path, Doc("Path to the pyproject.toml file.")],
    ) -> An[Config, Doc("Loaded configuration.")]:
        """Load configuration from pyproject.toml."""
        with open(path, "rb") as file:
            return cls.from_data(tomllib.load(file).get("tool", {}).get("yore", {}))

    @classmethod
    def from_default_locations(cls) -> An[Config, Doc("Loaded configuration.")]:
        """Load configuration from the default locations."""
        paths = ("config/yore.toml", "yore.toml", "pyproject.toml")
        cwd = Path.cwd()
        while True:
            for path in paths:
                if (cwd / path).exists():
                    if path == "pyproject.toml":
                        return cls.from_pyproject(cwd / path)
                    return cls.from_file(cwd / path)
            if cwd == cwd.parent:
                break
            cwd = cwd.parent
        return cls()



from __future__ import annotations

import json
import logging
import re
import subprocess
import sys
from dataclasses import dataclass
from datetime import date as Date  # noqa: N812
from datetime import datetime as DateTime  # noqa: N812
from datetime import timedelta as TimeDelta  # noqa: N812
from datetime import timezone as TimeZone  # noqa: N812
from functools import cache
from re import Pattern
from typing import TYPE_CHECKING, ClassVar, Literal
from urllib.request import urlopen

from humanize import naturaldelta
from packaging.version import Version

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

YoreKind = Literal["bump", "eol", "bol"]
"""The supported kinds of Yore comments."""

Scope = Literal["block", "file", "line"]
"""The scope of a comment."""

DEFAULT_PREFIX = "YORE"
"""The default prefix for Yore comments."""

DEFAULT_EXCLUDE = [".*", "__py*", "build", "dist"]
"""The default patterns to exclude when scanning directories."""

_logger = logging.getLogger(__name__)


def _indent(line: str) -> int:
    return len(line) - len(line.lstrip())


def _block_size(buffer: list[str], start: int) -> int:
    size = 0
    consecutive_blank = 0
    indent = _indent(buffer[start])
    for line in buffer[start:]:
        if line.strip():
            line_indent = _indent(line)
            if line_indent < indent:
                break
            if _indent(line) == indent and consecutive_blank:
                break
            consecutive_blank = 0
        else:
            consecutive_blank += 1
        size += 1
    return size - consecutive_blank


def _scope_range(replace: Scope, buffer: list[str], start: int) -> tuple[int, int]:
    if replace == "line":
        return start, start + 1
    if replace == "block":
        return start, start + _block_size(buffer, start)
    if replace == "file":
        return 0, len(buffer)
    raise ValueError(f"Invlid replace scope: {replace}")


def _reindent(lines: list[str], indent: int) -> list[str]:
    common = min(_indent(line) for line in lines)
    new = indent * " "
    return [f"{new}{line[common:]}" for line in lines]


def _match_to_line(match: re.Match) -> int | None:
    if matched_line := match.group("line"):
        return int(matched_line)
    return None


def _match_to_lines(match: re.Match) -> list[int] | None:
    if matched_lines := match.group("lines"):
        lines: list[int] = []
        matched_lines = matched_lines.replace(" ", ",").strip(",")
        matched_lines = re.sub(",+", ",", matched_lines)
        for line_range in matched_lines.split(","):
            if "-" in line_range:
                start, end = line_range.split("-")
                lines.extend(range(int(start), int(end) + 1))
            else:
                lines.append(int(line_range))
        return lines
    return None


def _match_to_comment(match: re.Match, file: Path, lineno: int) -> YoreComment:
    return YoreComment(
        file=file,
        lineno=lineno,
        raw=match.group(0),
        prefix=match.group("prefix"),
        suffix=match.group("suffix"),
        kind=match.group("kind"),
        version=match.group("version"),
        remove=match.group("remove"),
        replace=match.group("replace"),
        line=_match_to_line(match),
        lines=_match_to_lines(match),
        string=match.group("string"),
        regex=bool(match.group("regex")),
        pattern1=match.group("pattern1"),
        pattern2=match.group("pattern2"),
        within=match.group("within"),
    )


def _within(delta: TimeDelta, of: Date) -> bool:
    return DateTime.now(tz=TimeZone.utc).date() >= of - delta


def _delta(until: Date) -> TimeDelta:
    return until - DateTime.now(tz=TimeZone.utc).date()


# DUE: EOL 3.9: Remove block.
_dataclass_opts: dict[str, bool] = {}
if sys.version_info >= (3, 10):
    _dataclass_opts["kw_only"] = True


# DUE: EOL 3.9: Replace `**_dataclass_opts` with `kw_only=True` within line.
@dataclass(**_dataclass_opts)
class YoreComment:
    """A Yore comment."""

    file: Path
    """The file containing comment."""
    lineno: int
    """The line number of the comment."""
    raw: str
    """The raw comment."""
    prefix: str
    """The prefix of the comment."""
    suffix: str
    """The suffix of the comment."""
    kind: YoreKind
    """The kind of comment."""
    version: str
    """The EOL/bump version."""
    remove: Scope | None = None
    """The removal scope."""
    replace: Scope | None = None
    """The replacement scope."""
    line: int | None = None
    """The line to replace."""
    lines: list[int] | None = None
    """The lines to replace."""
    string: str | None = None
    """The string to replace."""
    regex: bool = False
    """Whether to use regex for replacement."""
    pattern1: str | None = None
    """The pattern to replace."""
    pattern2: str | None = None
    """The replacement pattern."""
    within: Scope | None = None
    """The scope to replace within."""

    @property
    def is_bol(self) -> bool:
        """Whether the comment is an End of Life comment."""
        return self.kind.lower() == "bol"

    @property
    def is_eol(self) -> bool:
        """Whether the comment is an End of Life comment."""
        return self.kind.lower() == "eol"

    @property
    def is_bump(self) -> bool:
        """Whether the comment is a bump comment."""
        return self.kind.lower() == "bump"

    @property
    def bol(self) -> Date:
        """The Beginning of Life date for the Python version."""
        return python_dates[self.version][0]

    @property
    def eol(self) -> Date:
        """The End of Life date for the Python version."""
        return python_dates[self.version][1]

    @property
    def comment(self) -> str:
        """The comment without the prefix."""
        return self.raw.removeprefix(self.prefix).removesuffix(self.suffix)

    def check(
        self,
        *,
        bump: str | None = None,
        eol_within: TimeDelta | None = None,
        bol_within: TimeDelta | None = None,
    ) -> bool:
        """Check the comment.

        Parameters:
            bump: The next version of the project.
            eol_within: The time delta to start warning before the End of Life of a Python version.
            bol_within: The time delta to start warning before the Beginning of Life of a Python version.

        Returns:
            True when there is nothing to do, False otherwise.
        """
        msg_location = f"{self.file}:{self.lineno}:"
        if self.is_eol:
            if eol_within and _within(eol_within, self.eol):
                _logger.warning(f"{msg_location} in ~{naturaldelta(_delta(self.eol))} {self.comment}")
            elif _within(TimeDelta(days=0), self.eol):
                _logger.error(f"{msg_location} since {self.eol} {self.comment}")
            else:
                return True
        elif self.is_bol:
            if bol_within and _within(bol_within, self.bol):
                _logger.warning(f"{msg_location} in ~{naturaldelta(_delta(self.bol))} {self.comment}")
            elif _within(TimeDelta(days=0), self.bol):
                _logger.error(f"{msg_location} since {self.bol} {self.comment}")
            else:
                return True
        elif self.is_bump and bump and Version(bump) >= Version(self.version):
            _logger.error(f"{msg_location} version {self.version} >= {self.comment}")
        else:
            return True
        return False

    def fix(
        self,
        buffer: list[str] | None = None,
        *,
        bump: str | None = None,
        eol_within: TimeDelta | None = None,
        bol_within: TimeDelta | None = None,
    ) -> bool:
        """Fix the comment and code below it.

        Parameters:
            buffer: The buffer to fix. If not provided, read from and write to the file.
            bump: The next version of the project.
            eol_within: The time delta to start fixing before the End of Life of a Python version.
            bol_within: The time delta to start fixing before the Beginning of Life of a Python version.

        Returns:
            Whether the comment was fixed.
        """
        write = buffer is None
        buffer = buffer or self.file.read_text().splitlines(keepends=True)

        # Check if the fix should be applied.
        if (
            (self.is_eol and ((eol_within and _within(eol_within, self.eol)) or _within(TimeDelta(days=0), self.eol)))
            or (
                self.is_bol and ((bol_within and _within(bol_within, self.bol)) or _within(TimeDelta(days=0), self.bol))
            )
            or (self.is_bump and bump and Version(bump) >= Version(self.version))
        ):
            # Start at the commnent line, immediately remove it.
            start = self.lineno - 1
            del buffer[start]

            if self.remove:
                start, end = _scope_range(self.remove, buffer, start)
                del buffer[start:end]
                if self.remove == "file":
                    self.file.unlink()

            elif self.replace:
                # Line numbers/ranges are relative to block starts, absolute for the "file" scope.
                start, end = _scope_range(self.replace, buffer, start)
                if self.line:
                    replacement = [buffer[start + self.line - 1]]
                elif self.lines:
                    replacement = [buffer[start + line] for line in self.lines]
                elif self.string:
                    replacement = [self.string + "\n"]
                else:
                    raise RuntimeError("No replacement specified")
                replacement = _reindent(replacement, _indent(buffer[start]))
                buffer[start:end] = replacement

            elif self.within:
                # Line numbers/ranges are relative to block starts, absolute for the "file" scope.
                start, end = _scope_range(self.within, buffer, start)
                block = buffer[start:end]
                if self.regex:
                    pattern1: Pattern = re.compile(self.pattern1)
                    replacement = [pattern1.sub(self.pattern2, line) for line in block]
                else:
                    replacement = [line.replace(self.pattern1, self.pattern2) for line in block]  # type: ignore[arg-type]
                replacement = _reindent(replacement, _indent(buffer[start]))
                buffer[start:end] = replacement

            if write and buffer:
                self.file.write_text("".join(buffer))

            return True
        return False


COMMENT_PREFIXES: set[str] = {
    r"\#\ ",  # Nim, Perl, PHP, Python, R, Ruby, shell, YAML
    r"//\ ",  # C, C++, Go, Java, Javascript, Rust, Swift
    r"--\ ",  # Haskell, Lua, SQL
    r";",  # Lisp, Scheme
    r"%\ ",  # MATLAB
    r"'\ ?",  # VBA
    r"/\*\ ",  # C, C++, Java, Javascript, CSS
    r"<!--\ ",  # HTML, Markdown, XML
    r"\{\#-?\ ",  # Jinja
    r"\(\*\ ",  # OCaml
}
"""The supported comment prefixes."""

_PATTERN_PREFIX = rf"^(?P<prefix>\s*(?:{'|'.join(sorted(COMMENT_PREFIXES))})PREFIX:\ )"
_PATTERN_SUFFIX = r"(?P<suffix>\.?.*)$"

COMMENT_PATTERN: str = r"""
    (?P<kind>bump|eol)\ (?P<version>[^:]+):\ (?:
        remove\ (?P<remove>block|file|line)
        |
        replace\ (?P<replace>block|file|line)\ with\ (?:
            line\ (?P<line>\d+)
            |
            lines\ (?P<lines>[\d, -]+)
            |
            `(?P<string>.+)`
        )
        |
        (?P<regex>regex-)?replace\ `(?P<pattern1>.+)`\ with\ `(?P<pattern2>.*)`\ within\ (?P<within>block|file|line)
    )
"""
"""The Yore comment pattern, as a regular expression."""


@cache
def get_pattern(prefix: str = DEFAULT_PREFIX) -> Pattern:
    """Get the Yore comment pattern with a specific prefix.

    Parameters:
        prefix: The prefix to use in the pattern.

    Returns:
        The Yore comment pattern.
    """
    return re.compile(
        _PATTERN_PREFIX.replace("PREFIX", prefix) + COMMENT_PATTERN + _PATTERN_SUFFIX,
        re.VERBOSE | re.IGNORECASE,
    )


@cache
def _get_prematching_pattern(prefix: str = DEFAULT_PREFIX) -> Pattern:
    return re.compile(_PATTERN_PREFIX.replace("PREFIX", prefix), re.VERBOSE | re.IGNORECASE)


def yield_files(directory: Path, exclude: list[str] | None = None) -> Iterator[Path]:
    """Yield all files in a directory."""
    exclude = DEFAULT_EXCLUDE if exclude is None else exclude
    _logger.debug(f"{directory}: scanning...")
    try:
        git_files = subprocess.run(  # noqa: S603
            ["git", "ls-files", "-z"],  # noqa: S607
            capture_output=True,
            cwd=directory,
            text=True,
            check=False,
        ).stdout
    except (FileNotFoundError, subprocess.CalledProcessError):
        for path in directory.iterdir():
            if path.is_file():
                yield path
            elif path.is_dir() and not any(path.match(pattern) for pattern in exclude):
                yield from yield_files(path, exclude=exclude)
    else:
        for filepath in git_files.strip("\0").split("\0"):
            yield directory / filepath


def yield_buffer_comments(file: Path, lines: list[str], *, prefix: str = DEFAULT_PREFIX) -> Iterator[YoreComment]:
    """Yield all Yore comments in a buffer.

    Parameters:
        file: The file to check.
        lines: The buffer to check (pre-read lines).
        prefix: The prefix to look for in the comments.

    Yields:
        Yore comments.
    """
    prepattern = _get_prematching_pattern(prefix)
    pattern = get_pattern(prefix)
    for lineno, line in enumerate(lines, 1):
        if prepattern.match(line):
            if match := pattern.match(line):
                yield _match_to_comment(match, file, lineno)
            else:
                _logger.error(f"{file}:{lineno}: invalid Yore comment")


def yield_file_comments(file: Path, *, prefix: str = DEFAULT_PREFIX) -> Iterator[YoreComment]:
    """Yield all Yore comments in a file.

    Parameters:
        file: The file to check.
        prefix: The prefix to look for in the comments.

    Yields:
        Yore comments.
    """
    try:
        lines = file.read_text().splitlines()
    except (OSError, UnicodeDecodeError):
        return
    yield from yield_buffer_comments(file, lines, prefix=prefix)


def yield_directory_comments(directory: Path, *, prefix: str = DEFAULT_PREFIX) -> Iterator[YoreComment]:
    """Yield all Yore comments in a directory.

    Parameters:
        directory: The directory to check.
        prefix: The prefix to look for in the comments.

    Yields:
        Yore comments.
    """
    for file in yield_files(directory):
        yield from yield_file_comments(file, prefix=prefix)


def yield_path_comments(path: Path, *, prefix: str = DEFAULT_PREFIX) -> Iterator[YoreComment]:
    """Yield all Yore comments in a file or directory.

    Parameters:
        path: The file or directory to check.
        prefix: The prefix to look for in the comments.

    Yields:
        Yore comments.
    """
    if path.is_dir():
        yield from yield_directory_comments(path, prefix=prefix)
    else:
        yield from yield_file_comments(path, prefix=prefix)


class _LazyPythonDates:
    EOL_DATA_URL = "https://raw.githubusercontent.com/python/devguide/main/include/release-cycle.json"
    _dates: ClassVar[dict[str, tuple[Date, Date]]] = {}

    def __getitem__(self, version: str) -> tuple[Date, Date]:
        if not self._dates:
            self._fetch()
        return self._dates[version]

    @staticmethod
    def _to_date(date: str) -> Date:
        parts = [int(part) for part in date.split("-")]
        if len(parts) == 2:  # noqa: PLR2004
            # Without a day, assume date to be the first of the next month.
            year, month = parts
            if month == 12:  # noqa: PLR2004
                month = 1
                year += 1
            else:
                month += 1
            day = 1
        else:
            year, month, day = parts
        return Date(year, month, day)

    def _fetch(self) -> None:
        data = json.loads(urlopen(self.EOL_DATA_URL).read())  # noqa: S310
        for version, info in data.items():
            bol_date = self._to_date(info["first_release"])
            eol_date = self._to_date(info["end_of_life"])
            self._dates[version] = (bol_date, eol_date)


python_dates = _LazyPythonDates()
"""A dictionary of Python versions and their Beginning/End of Life dates."""



"""Tests for the CLI."""

from __future__ import annotations

from typing import TYPE_CHECKING

from yore import main
from yore._internal import debug

if TYPE_CHECKING:
    import pytest


def test_main() -> None:
    """Basic CLI test."""
    assert main([]) == 2


def test_show_help(capsys: pytest.CaptureFixture) -> None:
    """Show help.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    assert main(["-h"]) == 0
    captured = capsys.readouterr()
    assert "yore" in captured.out


def test_show_version(capsys: pytest.CaptureFixture) -> None:
    """Show version.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    assert main(["-V"]) == 0
    captured = capsys.readouterr()
    assert debug._get_version() in captured.out


def test_show_debug_info(capsys: pytest.CaptureFixture) -> None:
    """Show debug information.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    assert main(["--debug-info"]) == 0
    captured = capsys.readouterr().out.lower()
    assert "python" in captured
    assert "system" in captured
    assert "environment" in captured
    assert "packages" in captured



"""Tests for the `cli` module."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from yore._internal import lib

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize(
    ("block", "expected_size"),
    [
        (["a", "b", "c"], 3),
        (["a", " b", "c"], 3),
        ([" a", " b", "c"], 2),
        (["a", "", "c"], 1),
        (["a", " b", "", "d"], 2),
        (["a", " b", "", " d"], 4),
        (["a", " b", "", " d", "e"], 5),
    ],
)
def test_block_size(block: list[str], expected_size: int) -> None:
    """Assert that `_block_size` returns the expected size."""
    assert lib._block_size(block, 0) == expected_size


class _Match:
    def __init__(self, lines: str) -> None:
        self.lines = lines

    def group(self, name: str) -> str:  # noqa: ARG002
        return self.lines


@pytest.mark.parametrize(
    ("lines", "expected_lines"),
    [
        ("1", [1]),
        ("1 2", [1, 2]),
        ("1,2", [1, 2]),
        (",, ,1,, ,,,,  2 ,,", [1, 2]),
        ("1-3", [1, 2, 3]),
        ("1-2", [1, 2]),
        ("1-1", [1]),
        ("1-2, 3, 5-7", [1, 2, 3, 5, 6, 7]),
    ],
)
def test_match_to_lines(lines: str, expected_lines: list[int]) -> None:
    """Assert that `_match_to_lines` returns the expected lines."""
    match = _Match(lines)
    assert lib._match_to_lines(match) == expected_lines  # type: ignore[arg-type]


def test_removing_file(tmp_path: Path) -> None:
    """Files are removed by "remove" comments and "file" scope."""
    file = tmp_path / "file1.py"
    file.write_text("# YORE: Bump 1: Remove file.")
    next(lib.yield_file_comments(file)).fix(bump="1")
    assert not file.exists()



# Why does this file exist, and why not put this in `__main__`?
#
# You might be tempted to import things from `__main__` later,
# but that will cause problems: the code will get executed twice:
#
# - When you run `python -m yore` python will execute
#   `__main__.py` as a script. That means there won't be any
#   `yore.__main__` in `sys.modules`.
# - When you import `__main__` it will get executed again (as a module) because
#   there's no `yore.__main__` in `sys.modules`.

from __future__ import annotations

import logging
import re
import subprocess
import sys
from dataclasses import dataclass, field
from datetime import timedelta
from difflib import unified_diff
from functools import wraps
from inspect import cleandoc
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable
from typing import Annotated as An

import cappa
from typing_extensions import Doc

from yore._internal import debug
from yore._internal.config import Config, Unset
from yore._internal.lib import DEFAULT_PREFIX, yield_buffer_comments, yield_files, yield_path_comments

if TYPE_CHECKING:
    from collections.abc import Iterator


_NAME = "yore"

_logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class _FromConfig(cappa.ValueFrom):
    def __init__(self, field: Unset | property, /) -> None:
        attr_name = field.fget.__name__ if isinstance(field, property) else field.name  # type: ignore[union-attr]
        super().__init__(self._from_config, attr_name=attr_name)

    @staticmethod
    def _from_config(attr_name: str) -> Any:
        # DUE: EOL 3.9: Replace `_load_config()` with `CommandMain._load_config()` within line.
        config = _load_config()
        value = getattr(config, attr_name)
        return cappa.Empty if isinstance(value, Unset) else value


def _parse_timedelta(value: str) -> timedelta:
    """Parse a timedelta from a string."""
    number, unit = re.match(r" *(\d+) *([a-z])[a-z]* *", value).groups()  # type: ignore[union-attr]
    multiplier = {"d": 1, "w": 7, "m": 31, "y": 365}[unit]
    return timedelta(days=int(number) * multiplier)


# DUE: EOL 3.9: Remove block.
_dataclass_opts: dict[str, bool] = {}
if sys.version_info >= (3, 10):
    _dataclass_opts["kw_only"] = True


@cappa.command(
    name="check",
    help="Check Yore comments.",
    description=cleandoc(
        """
        This command checks existing Yore comments in your code base
        against Python End of Life dates or the provided next version of your project.
        """,
    ),
)
# DUE: EOL 3.9: Replace `**_dataclass_opts` with `kw_only=True` within line.
@dataclass(**_dataclass_opts)
class CommandCheck:
    """Command to check Yore comments."""

    paths: An[
        list[Path],
        cappa.Arg(),
        Doc("Path to files or directories to check."),
    ] = field(default_factory=list)

    bump: An[
        str | None,
        cappa.Arg(short=True, long=True, value_name="VERSION"),
        Doc("The next version of your project."),
    ] = None

    eol_within: An[
        timedelta | None,
        cappa.Arg(short="-E", long="--eol/--eol-within", parse=_parse_timedelta, value_name="TIMEDELTA"),
        Doc(
            """
            The time delta to start checking before the End of Life of a Python version.
            It is provided in a human-readable format, like `2 weeks` or `1 month`.
            Spaces are optional, and the unit can be shortened to a single letter:
            `d` for days, `w` for weeks, `m` for months, and `y` for years.
            """,
        ),
    ] = None

    bol_within: An[
        timedelta | None,
        cappa.Arg(short="-B", long="--bol/--bol-within", parse=_parse_timedelta, value_name="TIMEDELTA"),
        Doc(
            """
            The time delta to start checking before the Beginning of Life of a Python version.
            It is provided in a human-readable format, like `2 weeks` or `1 month`.
            Spaces are optional, and the unit can be shortened to a single letter:
            `d` for days, `w` for weeks, `m` for months, and `y` for years.
            """,
        ),
    ] = None

    prefix: An[
        str,
        cappa.Arg(
            short="-p",
            long=True,
            num_args=1,
            default=_FromConfig(Config.prefix),
            show_default=f"{Config.prefix} or `{DEFAULT_PREFIX}`",
        ),
        Doc("""The prefix for Yore comments."""),
    ] = DEFAULT_PREFIX

    def __call__(self) -> int:
        """Check Yore comments."""
        ok = True
        paths = self.paths or [Path(".")]
        for path in paths:
            for comment in yield_path_comments(path, prefix=self.prefix):
                ok &= comment.check(bump=self.bump, eol_within=self.eol_within, bol_within=self.bol_within)
        return 0 if ok else 1


@cappa.command(
    name="diff",
    help="See the diff you would get after fixing comments.",
    description=cleandoc(
        """
        This command fixes all relevant Yore comments, then computes and prints
        a Git-like diff in the console.
        """,
    ),
)
# DUE: EOL 3.9: Replace `**_dataclass_opts` with `kw_only=True` within line.
@dataclass(**_dataclass_opts)
class CommandDiff:
    """Command to diff Yore comments."""

    paths: An[
        list[Path],
        cappa.Arg(),
        Doc("Path to files or directories to diff."),
    ] = field(default_factory=list)

    bump: An[
        str | None,
        cappa.Arg(short=True, long=True, value_name="VERSION"),
        Doc("The next version of your project."),
    ] = None

    eol_within: An[
        timedelta | None,
        cappa.Arg(short="-E", long="--eol/--eol-within", parse=_parse_timedelta, value_name="TIMEDELTA"),
        Doc(
            """
            The time delta to start diffing before the End of Life of a Python version.
            It is provided in a human-readable format, like `2 weeks` or `1 month`.
            Spaces are optional, and the unit can be shortened to a single letter:
            `d` for days, `w` for weeks, `m` for months, and `y` for years.
            """,
        ),
    ] = None

    bol_within: An[
        timedelta | None,
        cappa.Arg(short="-B", long="--bol/--bol-within", parse=_parse_timedelta, value_name="TIMEDELTA"),
        Doc(
            """
            The time delta to start diffing before the Beginning of Life of a Python version.
            It is provided in a human-readable format, like `2 weeks` or `1 month`.
            Spaces are optional, and the unit can be shortened to a single letter:
            `d` for days, `w` for weeks, `m` for months, and `y` for years.
            """,
        ),
    ] = None

    highlight: An[
        str | None,
        cappa.Arg(
            short="-H",
            long="--highlight",
            num_args=1,
            default=_FromConfig(Config.diff_highlight),
            show_default=f"{Config.diff_highlight}",
        ),
        Doc("The command to highlight diffs."),
    ] = None

    prefix: An[
        str,
        cappa.Arg(
            short="-p",
            long=True,
            num_args=1,
            default=_FromConfig(Config.prefix),
            show_default=f"{Config.prefix} or `{DEFAULT_PREFIX}`",
        ),
        Doc("""The prefix for Yore comments."""),
    ] = DEFAULT_PREFIX

    def _diff(self, file: Path) -> Iterator[str]:
        try:
            old_lines = file.read_text().splitlines(keepends=True)
        except (OSError, UnicodeDecodeError):
            return
        new_lines = old_lines.copy()
        for comment in sorted(
            yield_buffer_comments(file, new_lines, prefix=self.prefix),
            key=lambda c: c.lineno,
            reverse=True,
        ):
            comment.fix(buffer=new_lines, bump=self.bump, eol_within=self.eol_within, bol_within=self.bol_within)
        yield from unified_diff(old_lines, new_lines, fromfile=str(file), tofile=str(file))

    def _diff_paths(self, paths: list[Path]) -> Iterator[str]:
        for path in paths:
            if path.is_file():
                yield from self._diff(path)
            else:
                for file in sorted(yield_files(path)):
                    yield from self._diff(file)

    def __call__(self) -> int:
        """Diff Yore comments."""
        lines = self._diff_paths(self.paths or [Path(".")])
        if self.highlight:
            process = subprocess.Popen(self.highlight, shell=True, text=True, stdin=subprocess.PIPE)  # noqa: S602
            for line in lines:
                process.stdin.write(line)  # type: ignore[union-attr]
            process.stdin.close()  # type: ignore[union-attr]
            process.wait()
            return int(process.returncode)
        for line in lines:
            print(line, end="")
        return 0


@cappa.command(
    name="fix",
    help="Fix Yore comments and the associated code lines.",
    description=cleandoc(
        """
        This command will fix your code by transforming it according to the Yore comments.
        """,
    ),
)
# DUE: EOL 3.9: Replace `**_dataclass_opts` with `kw_only=True` within line.
@dataclass(**_dataclass_opts)
class CommandFix:
    """Command to fix Yore comments."""

    paths: An[
        list[Path],
        cappa.Arg(),
        Doc("Path to files or directories to fix."),
    ] = field(default_factory=list)

    bump: An[
        str | None,
        cappa.Arg(short=True, long=True, value_name="VERSION"),
        Doc("The next version of your project."),
    ] = None

    eol_within: An[
        timedelta | None,
        cappa.Arg(short="-E", long="--eol/--eol-within", parse=_parse_timedelta, value_name="TIMEDELTA"),
        Doc(
            """
            The time delta to start fixing before the End of Life of a Python version.
            It is provided in a human-readable format, like `2 weeks` or `1 month`.
            Spaces are optional, and the unit can be shortened to a single letter:
            `d` for days, `w` for weeks, `m` for months, and `y` for years.
            """,
        ),
    ] = None

    bol_within: An[
        timedelta | None,
        cappa.Arg(short="-B", long="--bol/--bol-within", parse=_parse_timedelta, value_name="TIMEDELTA"),
        Doc(
            """
            The time delta to start fixing before the Beginning of Life of a Python version.
            It is provided in a human-readable format, like `2 weeks` or `1 month`.
            Spaces are optional, and the unit can be shortened to a single letter:
            `d` for days, `w` for weeks, `m` for months, and `y` for years.
            """,
        ),
    ] = None

    prefix: An[
        str,
        cappa.Arg(
            short="-p",
            long=True,
            num_args=1,
            default=_FromConfig(Config.prefix),
            show_default=f"{Config.prefix}",
        ),
        Doc("""The prefix for Yore comments."""),
    ] = DEFAULT_PREFIX

    def _fix(self, file: Path) -> None:
        try:
            lines = file.read_text().splitlines(keepends=True)
        except (OSError, UnicodeDecodeError):
            return
        count = 0
        for comment in sorted(
            yield_buffer_comments(file, lines, prefix=self.prefix),
            key=lambda c: c.lineno,
            reverse=True,
        ):
            if comment.fix(buffer=lines, bump=self.bump, eol_within=self.eol_within, bol_within=self.bol_within):
                count += 1
        if count:
            file.write_text("".join(lines))
            _logger.info(f"fixed {count} comment{'s' if count > 1 else ''} in {file}")

    def __call__(self) -> int:
        """Fix Yore comments."""
        paths = self.paths or [Path(".")]
        for path in paths:
            if path.is_file():
                self._fix(path)
            else:
                for file in yield_files(path):
                    self._fix(file)
        return 0


# DUE: EOL 3.9: Remove block.
def _print_and_exit(
    func: An[Callable[[], str | None], Doc("A function that returns or prints a string.")],
    code: An[int, Doc("The status code to exit with.")] = 0,
) -> Callable[[], None]:
    """Argument action callable to print something and exit immediately."""

    @wraps(func)
    def _inner() -> None:
        raise cappa.Exit(func() or "", code=code)

    return _inner


# DUE: EOL 3.9: Remove block.
def _load_config(file: Path | None = None) -> Config:
    if CommandMain._CONFIG is None:
        CommandMain._CONFIG = Config.from_file(file) if file else Config.from_default_locations()
    return CommandMain._CONFIG


@cappa.command(
    name=_NAME,
    help="Manage legacy code in your code base with YORE comments.",
    description=cleandoc(
        """
        This tool lets you add `# YORE` comments (similar to `# TODO` comments)
        that will help you manage legacy code in your code base.

        A YORE comment follows a simple syntax that tells why this legacy code
        is here and how it can be checked, or fixed once it's time to do so.
        The syntax is as follows:

        ```python
        # <PREFIX>: <eol|bump> <VERSION>: Remove <block|line>.
        # <PREFIX>: <eol|bump> <VERSION>: replace <block|line> with line <LINENO>.
        # <PREFIX>: <eol|bump> <VERSION>: replace <block|line> with lines <LINE-RANGE1[, LINE-RANGE2...]>.
        # <PREFIX>: <eol|bump> <VERSION>: replace <block|line> with `<STRING>`.
        # <PREFIX>: <eol|bump> <VERSION>: [regex-]replace `<PATTERN1>` with `<PATTERN2>` within <block|line>.
        ```

        Terms between `<` and `>` *must* be provided, while terms between `[` and `]` are optional.
        Uppercase terms are placeholders that you should replace with actual values,
        while lowercase terms are keywords that you should use literally.
        Everything except placeholders is case-insensitive.

        The default **prefix** is `YORE`. For now it is only configurable through the Python API.

        Examples:

        *Replace a block of code when Python 3.8 reaches its End of Life.
        In this example, we want to replace the block with `from ast import unparse`.*

        ```python
        # YORE: EOL 3.8: Replace block with line 4.
        if sys.version_info < (3, 9):
            from astunparse import unparse
        else:
            from ast import unparse
        ```

        *Replace `lstrip` by `removeprefix` when Python 3.8 reaches its End of Life.*

        ```python
        # YORE: EOL 3.8: Replace `lstrip` with `removeprefix` within line.
        return [cpn.lstrip("_") for cpn in a.split(".")] == [cpn.lstrip("_") for cpn in b.split(".")]
        ```

        *Simplify union of accepted types when we bump the project to version 1.0.0.*

        ```python
        def load_extensions(
            # YORE: Bump 1.0.0: Replace ` | Sequence[LoadableExtension],` with `` within line.
            *exts: LoadableExtension | Sequence[LoadableExtension],
        ): ...
        ```
        """,
    ),
)
# DUE: EOL 3.9: Replace `**_dataclass_opts` with `kw_only=True` within line.
@dataclass(**_dataclass_opts)
class CommandMain:
    """Command to manage legacy code in your code base with YORE comments."""

    subcommand: An[cappa.Subcommands[CommandCheck | CommandDiff | CommandFix], Doc("The selected subcommand.")]

    # DUE: EOL 3.9: Replace `# ` with `` within block.
    # @staticmethod
    # def _load_config(file: Path | None = None) -> Config:
    #     if CommandMain._CONFIG is None:
    #         CommandMain._CONFIG = Config.from_file(file) if file else Config.from_default_locations()
    #     return CommandMain._CONFIG

    # DUE: EOL 3.9: Replace `# ` with `` within block.
    # @staticmethod
    # def _print_and_exit(
    #     func: An[Callable[[], str | None], Doc("A function that returns or prints a string.")],
    #     code: An[int, Doc("The status code to exit with.")] = 0,
    # ) -> Callable[[], None]:
    #     """Argument action callable to print something and exit immediately."""
    #
    #     @wraps(func)
    #     def _inner() -> None:
    #         raise cappa.Exit(func() or "", code=code)
    #
    #     return _inner

    # DUE: EOL 3.9: Regex-replace `Config \| None = .*` with `ClassVar[Config | None] = None` within line.
    _CONFIG: Config | None = field(default=None, init=False, repr=False)

    config: An[
        Config,
        cappa.Arg(
            short="-c",
            long=True,
            parse=_load_config,
            propagate=True,
            show_default="`config/yore.toml`, `yore.toml`, or `pyproject.toml`",
        ),
        Doc("Path to the configuration file."),
    ] = field(default_factory=_load_config)

    version: An[
        bool,
        cappa.Arg(
            short="-V",
            long=True,
            action=_print_and_exit(debug._get_version),
            num_args=0,
            help="Print the program version and exit.",
        ),
        Doc("Version CLI option."),
    ] = False

    debug_info: An[
        bool,
        cappa.Arg(long=True, action=_print_and_exit(debug._print_debug_info), num_args=0),
        Doc("Print debug information."),
    ] = False


def main(
    args: An[list[str] | None, Doc("Arguments passed from the command line.")] = None,
) -> An[int, Doc("An exit code.")]:
    """Run the main program.

    This function is executed when you type `yore` or `python -m yore`.
    """
    logging.basicConfig(level=logging.INFO, format="%(message)s")
    output = cappa.Output(error_format=f"[bold]{_NAME}[/]: [bold red]error[/]: {{message}}")
    completion_option: cappa.Arg = cappa.Arg(
        long=True,
        action=cappa.ArgAction.completion,
        choices=["complete", "generate"],
        help="Print shell-specific completion source.",
    )
    help_option: cappa.Arg = cappa.Arg(
        short="-h",
        long=True,
        action=cappa.ArgAction.help,
        help="Print the program help and exit.",
    )
    help_formatter = cappa.HelpFormatter(default_format="Default: {default}.")

    try:
        return cappa.invoke(
            CommandMain,
            argv=args,
            output=output,
            help=help_option,
            completion=completion_option,
            help_formatter=help_formatter,
        )
    except cappa.Exit as exit:
        return int(1 if exit.code is None else exit.code)



"""yore package.

Manage legacy code with comments.
"""

from __future__ import annotations

from yore._internal.cli import CommandCheck, CommandDiff, CommandFix, CommandMain, main
from yore._internal.config import Config, Unset, config_field
from yore._internal.lib import (
    COMMENT_PATTERN,
    COMMENT_PREFIXES,
    DEFAULT_EXCLUDE,
    DEFAULT_PREFIX,
    Scope,
    YoreComment,
    YoreKind,
    get_pattern,
    python_dates,
    yield_buffer_comments,
    yield_directory_comments,
    yield_file_comments,
    yield_files,
    yield_path_comments,
)

__all__: list[str] = [
    "COMMENT_PATTERN",
    "COMMENT_PREFIXES",
    "DEFAULT_EXCLUDE",
    "DEFAULT_PREFIX",
    "CommandCheck",
    "CommandDiff",
    "CommandFix",
    "CommandMain",
    "Config",
    "Scope",
    "Unset",
    "YoreComment",
    "YoreKind",
    "config_field",
    "get_pattern",
    "main",
    "python_dates",
    "yield_buffer_comments",
    "yield_directory_comments",
    "yield_file_comments",
    "yield_files",
    "yield_path_comments",
]
